@@ -70,6 +70,7 @@ FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );
 #define ERRFILE_ansicoldef	       ( ERRFILE_CORE | 0x001e0000 )
 #define ERRFILE_fault		       ( ERRFILE_CORE | 0x001f0000 )
 #define ERRFILE_blocktrans	       ( ERRFILE_CORE | 0x00200000 )
+#define ERRFILE_pixbuf		       ( ERRFILE_CORE | 0x00210000 )
 
 #define ERRFILE_eisa		     ( ERRFILE_DRIVER | 0x00000000 )
 #define ERRFILE_isa		     ( ERRFILE_DRIVER | 0x00010000 )
@@ -85,6 +86,7 @@ FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );
 #define ERRFILE_uhci		     ( ERRFILE_DRIVER | 0x000b0000 )
 #define ERRFILE_usbhid		     ( ERRFILE_DRIVER | 0x000c0000 )
 #define ERRFILE_usbkbd		     ( ERRFILE_DRIVER | 0x000d0000 )
+#define ERRFILE_usbio		     ( ERRFILE_DRIVER | 0x000e0000 )
 
 #define ERRFILE_nvs		     ( ERRFILE_DRIVER | 0x00100000 )
 #define ERRFILE_spi		     ( ERRFILE_DRIVER | 0x00110000 )
@@ -182,6 +184,15 @@ FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );
 #define ERRFILE_smsc75xx	     ( ERRFILE_DRIVER | 0x00770000 )
 #define ERRFILE_intelvf		     ( ERRFILE_DRIVER | 0x00780000 )
 #define ERRFILE_intelxvf	     ( ERRFILE_DRIVER | 0x00790000 )
+#define ERRFILE_smsc95xx	     ( ERRFILE_DRIVER | 0x007a0000 )
+#define ERRFILE_acm		     ( ERRFILE_DRIVER | 0x007b0000 )
+#define ERRFILE_eoib		     ( ERRFILE_DRIVER | 0x007c0000 )
+#define ERRFILE_golan		     ( ERRFILE_DRIVER | 0x007d0000 )
+#define ERRFILE_flexboot_nodnic	     ( ERRFILE_DRIVER | 0x007e0000 )
+#define ERRFILE_virtio_pci	     ( ERRFILE_DRIVER | 0x007f0000 )
+#define ERRFILE_pciea		     ( ERRFILE_DRIVER | 0x00c00000 )
+#define ERRFILE_axge		     ( ERRFILE_DRIVER | 0x00c10000 )
+#define ERRFILE_thunderx	     ( ERRFILE_DRIVER | 0x00c20000 )
 
 #define ERRFILE_aoe			( ERRFILE_NET | 0x00000000 )
 #define ERRFILE_arp			( ERRFILE_NET | 0x00010000 )
@@ -255,6 +266,8 @@ FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );
 #define ERRFILE_peerdisc		( ERRFILE_NET | 0x00450000 )
 #define ERRFILE_peerblk			( ERRFILE_NET | 0x00460000 )
 #define ERRFILE_peermux			( ERRFILE_NET | 0x00470000 )
+#define ERRFILE_xsigo			( ERRFILE_NET | 0x00480000 )
+#define ERRFILE_ntp			( ERRFILE_NET | 0x00490000 )
 
 #define ERRFILE_image		      ( ERRFILE_IMAGE | 0x00000000 )
 #define ERRFILE_elf		      ( ERRFILE_IMAGE | 0x00010000 )
@@ -264,6 +277,8 @@ FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );
 #define ERRFILE_embedded	      ( ERRFILE_IMAGE | 0x00050000 )
 #define ERRFILE_pnm		      ( ERRFILE_IMAGE | 0x00060000 )
 #define ERRFILE_png		      ( ERRFILE_IMAGE | 0x00070000 )
+#define ERRFILE_der		      ( ERRFILE_IMAGE | 0x00080000 )
+#define ERRFILE_pem		      ( ERRFILE_IMAGE | 0x00090000 )
 
 #define ERRFILE_asn1		      ( ERRFILE_OTHER | 0x00000000 )
 #define ERRFILE_chap		      ( ERRFILE_OTHER | 0x00010000 )
@@ -339,6 +354,12 @@ FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );
 #define ERRFILE_vmbus		      ( ERRFILE_OTHER | 0x00470000 )
 #define ERRFILE_efi_time	      ( ERRFILE_OTHER | 0x00480000 )
 #define ERRFILE_efi_watchdog	      ( ERRFILE_OTHER | 0x00490000 )
+#define ERRFILE_efi_pxe		      ( ERRFILE_OTHER | 0x004a0000 )
+#define ERRFILE_efi_usb		      ( ERRFILE_OTHER | 0x004b0000 )
+#define ERRFILE_efi_fbcon	      ( ERRFILE_OTHER | 0x004c0000 )
+#define ERRFILE_efi_local	      ( ERRFILE_OTHER | 0x004d0000 )
+#define ERRFILE_efi_entropy	      ( ERRFILE_OTHER | 0x004e0000 )
+#define ERRFILE_cert_cmd	      ( ERRFILE_OTHER | 0x004f0000 )
 
 /** @} */
 
